# -*- coding: utf-8 -*-
from hao import places


def get_provinces():
    return [place.name for place in places.PLACES]


def get_cities():
    city_list = []
    for place in places.PLACES:
        if place.children is None or place.children_type is None or place.children_type != places.TYPE_CITY:
            continue

        for child in place.children:
            city = child.name
            city_list.append(city)
    return city_list


def get_provinces_and_cities():
    return get_provinces() + get_cities()



# -*- coding: utf-8 -*-
import abc
import importlib
import inspect
import os
from typing import Dict, Iterator, List, Optional, Union

import hao
import torch
import torch.nn as nn
from hao.namespaces import attr, from_args
from hao.stopwatch import Stopwatch
from peft import get_peft_config, get_peft_model, set_peft_model_state_dict
from torch.nn import DataParallel
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data.dataloader import default_collate
from transformers import PreTrainedModel, PreTrainedTokenizerFast

import tailors
from tailors.domains import Example, Factor, Tags
from tailors.exceptions import TailorsError

LOGGER = hao.logs.get_logger(__name__)


@from_args
class TailorsConf:
    meta: dict = None
    freeze_embedding: bool = attr(bool, default=False)
    seq_len = attr(int, default=128)


class TailorsIO(abc.ABC):
    def __init__(self, model_conf: TailorsConf, tokenizer: PreTrainedTokenizerFast) -> None:
        self.model_conf = model_conf
        self.tokenizer = tokenizer
        self.tags: Tags = self.build_tags()
        self.empty_sample = self.build_empty_sample()
        self.init()
        self.empty_input = self.build_empty_input()

    @abc.abstractmethod
    def build_tags(self) -> Tags:
        raise NotImplementedError()

    def build_empty_sample(self) -> Tags:
        return self.encode('', self.tokenizer, self.model_conf.seq_len)

    def init(self):
        pass

    @abc.abstractmethod
    def build_empty_input(self):
        raise NotImplementedError()

    @staticmethod
    def encode(text_or_tokens: Union[str, List[str]],
               tokenizer: PreTrainedTokenizerFast,
               seq_len: int,
               add_special_tokens=True,
               padding: Union[bool, str] = 'max_length',
               truncation: Union[bool, str] = True,
               return_offsets_mapping: bool = True,
               return_tensors: Optional[str] = 'pt'):
        is_text = isinstance(text_or_tokens, str)
        if is_text:
            encoder = tokenizer.encode_plus
        else:
            encoder = tokenizer.prepare_for_model
            if text_or_tokens and isinstance(text_or_tokens[0], str):
                text_or_tokens = tokenizer.convert_tokens_to_ids(text_or_tokens)
        encoded = encoder(
            text_or_tokens,
            add_special_tokens=add_special_tokens,
            max_length=seq_len,
            padding=padding,
            truncation=truncation,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_offsets_mapping=return_offsets_mapping,
            return_tensors=return_tensors,
        )
        input_ids = encoded.get("input_ids")
        attention_mask = encoded.get("attention_mask")
        token_type_ids = encoded.get("token_type_ids")

        if return_tensors:
            input_ids = input_ids.squeeze()
            attention_mask = attention_mask.squeeze()
            token_type_ids = token_type_ids.squeeze()

        if is_text and return_offsets_mapping:
            offset_mapping = encoded.get("offset_mapping")
            if return_tensors:
                offset_mapping = offset_mapping[0].tolist() if offset_mapping is not None else None
            return input_ids, attention_mask, token_type_ids, offset_mapping
        else:
            return input_ids, attention_mask, token_type_ids

    def from_line(self, line: str) -> Union[Example, List[Example]]:
        raise NotImplementedError()

    @abc.abstractmethod
    def for_inference(self, lines: List[str], bz: int = 32, *args, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def post_inference(self, *args, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def for_eval(self, lines: Iterator[str], bz: int = 32, *args, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def collate(batch):
        try:
            return default_collate(batch)
        except RuntimeError as e:
            LOGGER.exception(e)


class Tailors(nn.Module, abc.ABC):

    def __init__(self, model_conf: TailorsConf):
        super().__init__()
        self.device = None
        assert model_conf.meta is not None, '`meta` must be populated manually'
        self.model_conf = model_conf

        self.io: TailorsIO = self.get_io()
        self.plm = self.build_plm()

    def freeze(self):
        tailors.freeze(self)
        self.eval()

    def unfreeze(self) -> None:
        tailors.unfreeze(self)
        self.train()

    def use_device(self, device):
        if device:
            self.to(device)
            self.device = device
        return self

    @abc.abstractmethod
    def build_tokenizer(self) -> PreTrainedTokenizerFast:
        raise NotImplementedError()

    @abc.abstractmethod
    def build_plm(self) -> PreTrainedModel:
        raise NotImplementedError()

    def on_save_checkpoint(self):
        pass

    def get_io(self):
        class_name = f"{self.__class__.__name__}IO"
        module_name = self.__class__.__module__
        try:
            module = importlib.import_module(module_name)
            io_class = getattr(module, class_name)
        except AttributeError:
            raise TailorsError(f"[io] expecting io class: `{module_name}.{class_name}`")
        except Exception as e:
            raise TailorsError(f"[io] failed to init io class: `{module_name}.{class_name}`", e)
        tokenizer = self.build_tokenizer()
        return io_class(self.model_conf, tokenizer)

    @tailors.auto_device
    def forward(self, *args, **kwargs):
        features, = args
        logits, mask = self.encode(features)
        return self.decode(logits, mask)

    @abc.abstractmethod
    def encode(self, *args, **kwargs):
        """Make sure always return logits as the only/first value"""
        raise NotImplementedError()

    @abc.abstractmethod
    def decode(self, *args, **kwargs):
        raise NotImplementedError()

    def lr_factors(self) -> Dict[str, Factor]:
        return {
            'embedding': Factor(factor=0.1, max_val=1e-4),
            'crf': Factor(factor=1000, max_val=5e-2),
        }

    @classmethod
    def from_pretrained(cls, path_or_key: str, use_gpu = True):
        if inspect.isabstract(cls):
            raise TailorsError(f"Not supported call from abstract class: {cls.__name__}")

        if ".local" in path_or_key:  # do not use SEQUE_CONFIG, since it's not called in this lib project
            hao.oss.init(path_or_key[: path_or_key.rfind(".")])
            model_path = hao.config.get(path_or_key)
        else:
            model_path = path_or_key
        fullpath = hao.paths.get(model_path)

        if model_path is None or not os.path.isfile(fullpath):
            raise TailorsError(f"model not found: {model_path}")

        LOGGER.info(f"[{cls.__name__}] loading from: {model_path}")
        sw = Stopwatch()
        state_dicts = torch.load(fullpath)
        state_dict = state_dicts.get('state_dict')
        model = cls(state_dicts.get('model_conf'))
        if (peft_config := state_dicts.get('peft_config')) is None:
            model.load_state_dict(state_dict)
        else:
            model = get_peft_model(model, get_peft_config(peft_config))
            set_peft_model_state_dict(model, state_dict)

        model.freeze()
        LOGGER.info(f"[{cls.__name__}] loaded, took: {sw.took()}")

        if use_gpu:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            model.use_device(device)
        return model

    @abc.abstractmethod
    def predict(self, *args, **kwargs):
        raise NotImplementedError()

    def export_to_model(self, output_path):
        is_dp_module = isinstance(self, (DistributedDataParallel, DataParallel))
        model = self.module if is_dp_module else self

        model.on_save_checkpoint()
        state_dicts = {'state_dict': model.state_dict(), 'model_conf': model.model_conf}
        hao.paths.make_parent_dirs(output_path)
        torch.save(state_dicts, output_path)

    def export_to_onnx(self, output_path):
        torch.onnx.export(
            self,
            self.io.empty_input,
            output_path,
            verbose=False,
            input_names=['input_ids', 'attention_mask', 'token_type_id'],
            output_names=['tag', 'score'],
            dynamic_axes={
                'input_ids': {0: 'batch_size', 1: 'sequence'},  # 第0维是batch dimension
                'attention_mask': {0: 'batch_size', 1: 'sequence'},  # 第0维是batch dimension
                'token_type_id': {0: 'batch_size', 1: 'sequence'},  # 第0维是batch dimension
                'tag': {0: 'batch_size', 1: 'sequence'},
                'score': {0: 'batch_size', 1: 'sequence'},
            },
            export_params=True,
            opset_version=13,
            do_constant_folding=True,
        )
        LOGGER.info(f'saved onnx to: {output_path}')

    @classmethod
    def to_model(cls, model_path, output_path=None):
        if model_path is None:
            raise TailorsError('empty output_path')
        model_path = hao.paths.get_path(model_path)
        if not os.path.exists(model_path):
            raise TailorsError(f'model_path not exist: {model_path}')

        model = cls.from_pretrained(model_path)
        if output_path is None:
            path_base, _ = os.path.splitext(os.path.basename(model_path))
            output_path = hao.paths.get_path('data', 'model', f"{path_base}.bin")
        hao.paths.make_parent_dirs(output_path)
        model.export_to_model(output_path)
        return output_path

    @classmethod
    def to_onnx(cls, model_path, output_path=None):
        if model_path is None:
            raise TailorsError('empty output_path')
        model_path = hao.paths.get_path(model_path)
        if not os.path.exists(model_path):
            raise TailorsError(f'model_path not exist: {model_path}')

        model = cls.from_pretrained(model_path)
        if output_path is None:
            path_base, _ = os.path.splitext(os.path.basename(model_path))
            output_path = hao.paths.get_path('data', 'model', f"{path_base}.onnx")
        hao.paths.make_parent_dirs(output_path)
        model.export_to_onnx(output_path)
        return output_path



# -*- coding: utf-8 -*-
from typing import Optional

import torch
import torch.nn as nn
from torch import Tensor
from torch.autograd import Variable
from torch.nn import functional as F

from tailors.exceptions import TailorsError


class FocalLoss(nn.Module):
    def __init__(self, gamma=0, alpha=None, size_average=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float, int)):
            self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list):
            self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, logits, target):
        if logits.dim() > 2:
            logits = logits.view(logits.size(0), logits.size(1), -1)  # N,C,H,W => N,C,H*W
            logits = logits.transpose(1, 2)  # N,C,H*W => N,H*W,C
            logits = logits.contiguous().view(-1, logits.size(2))  # N,H*W,C => N*H*W,C
        target = target.view(-1, 1)

        logits = logits.log_softmax(1)
        logits = logits.gather(1, target)
        logits = logits.view(-1)
        pt = Variable(logits.data.exp())

        if self.alpha is not None:
            if self.alpha.type() != logits.data.type():
                self.alpha = self.alpha.type_as(logits.data)
            at = self.alpha.gather(0, target.data.view(-1))
            logits = logits * Variable(at)

        loss = -1 * (1 - pt) ** self.gamma * logits
        if self.size_average:
            return loss.mean()
        else:
            return loss.sum()


class MultiClassFocalLossWithAlpha(nn.Module):
    def __init__(self, alpha=[0.2, 0.3, 0.5], gamma=2, reduction='mean'):
        """
        :param alpha: 权重系数列表，三分类中第0类权重0.2，第1类权重0.3，第2类权重0.5
        :param gamma: 困难样本挖掘的gamma
        :param reduction:
        """
        super(MultiClassFocalLossWithAlpha, self).__init__()
        self.alpha = torch.tensor(alpha)
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, pred, target):
        alpha = self.alpha[target]  # 为当前batch内的样本，逐个分配类别权重，shape=(bs), 一维向量
        log_softmax = torch.log_softmax(pred, dim=1) # 对模型裸输出做softmax再取log, shape=(bs, 3)
        logpt = torch.gather(log_softmax, dim=1, index=target.view(-1, 1))  # 取出每个样本在类别标签位置的log_softmax值, shape=(bs, 1)
        logpt = logpt.view(-1)  # 降维，shape=(bs)
        ce_loss = -logpt  # 对log_softmax再取负，就是交叉熵了
        pt = torch.exp(logpt)  #对log_softmax取exp，把log消了，就是每个样本在类别标签位置的softmax值了，shape=(bs)
        focal_loss = alpha * (1 - pt) ** self.gamma * ce_loss  # 根据公式计算focal loss，得到每个样本的loss值，shape=(bs)
        if self.reduction == "mean":
            return torch.mean(focal_loss)
        if self.reduction == "sum":
            return torch.sum(focal_loss)
        return focal_loss


class DiceLoss(nn.Module):
    """
    Dice coefficient for short, is an F1-oriented statistic used to gauge the similarity of two sets.
    Given two sets A and B, the vanilla dice coefficient between them is given as follows:
        Dice(A, B)  = 2 * True_Positive / (2 * True_Positive + False_Positive + False_Negative)
                    = 2 * |A and B| / (|A| + |B|)
    Math Function:
        U-NET: https://arxiv.org/abs/1505.04597.pdf
        dice_loss(p, y) = 1 - numerator / denominator
            numerator = 2 * \sum_{1}^{t} p_i * y_i + smooth
            denominator = \sum_{1}^{t} p_i + \sum_{1} ^{t} y_i + smooth
        if square_denominator is True, the denominator is \sum_{1}^{t} (p_i ** 2) + \sum_{1} ^{t} (y_i ** 2) + smooth
        V-NET: https://arxiv.org/abs/1606.04797.pdf
    Args:
        smooth (float, optional): a manual smooth value for numerator and denominator.
        square_denominator (bool, optional): [True, False], specifies whether to square the denominator in the loss function.
        with_logits (bool, optional): [True, False], specifies whether the input tensor is normalized by Sigmoid/Softmax funcs.
        ohem_ratio: max ratio of positive/negative, defautls to 0.0, which means no ohem.
        alpha: dsc alpha
    Shape:
        - logits: (*)
        - target: (*)
        - mask: (*) 0,1 mask for the input sequence.
        - Output: Scalar loss
    Examples:
        >>> loss = DiceLoss(with_logits=True, ohem_ratio=0.1)
        >>> logits = torch.FloatTensor([2, 1, 2, 2, 1])
        >>> logits.requires_grad=True
        >>> target = torch.LongTensor([0, 1, 0, 0, 0])
        >>> output = loss(logits, target)
        >>> output.backward()
    """

    def __init__(self,
                 smooth: Optional[float] = 1e-4,
                 square_denominator: Optional[bool] = False,
                 with_logits: Optional[bool] = True,
                 ohem_ratio: float = 0.0,
                 alpha: float = 0.0,
                 reduction: Optional[str] = "mean",
                 index_label_position=True) -> None:
        super(DiceLoss, self).__init__()

        self.reduction = reduction
        self.with_logits = with_logits
        self.smooth = smooth
        self.square_denominator = square_denominator
        self.ohem_ratio = ohem_ratio
        self.alpha = alpha
        self.index_label_position = index_label_position

    def forward(self, logits: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        logits_size = logits.shape[-1]

        if logits_size != 1:
            loss = self._multiple_class(logits, target, logits_size, mask=mask)
        else:
            loss = self._binary_class(logits, target, mask=mask)

        if self.reduction == "mean":
            return loss.mean()
        if self.reduction == "sum":
            return loss.sum()
        return loss

    def _compute_dice_loss(self, flat_input, flat_target):
        flat_input = ((1 - flat_input) ** self.alpha) * flat_input
        intersection = torch.sum(flat_input * flat_target, -1)
        if not self.square_denominator:
            loss = 1 - ((2 * intersection + self.smooth) /
                        (flat_input.sum() + flat_target.sum() + self.smooth))
        else:
            loss = 1 - ((2 * intersection + self.smooth) /
                        (torch.sum(torch.square(flat_input, ), -1) + torch.sum(torch.square(flat_target), -1) + self.smooth))

        return loss

    def _multiple_class(self, logits, target, logits_size, mask=None):
        flat_input = logits
        flat_target = F.one_hot(target, num_classes=logits_size).float() if self.index_label_position else target.float()
        flat_input = torch.nn.Softmax(dim=1)(flat_input) if self.with_logits else flat_input

        if mask is not None:
            mask = mask.float()
            flat_input = flat_input * mask
            flat_target = flat_target * mask
        else:
            mask = torch.ones_like(target)

        loss = None
        if self.ohem_ratio > 0:
            mask_neg = torch.logical_not(mask)
            for label_idx in range(logits_size):
                pos_example = target == label_idx
                neg_example = target != label_idx

                pos_num = pos_example.sum()
                neg_num = mask.sum() - (pos_num - (mask_neg & pos_example).sum())
                keep_num = min(int(pos_num * self.ohem_ratio / logits_size), neg_num)

                if keep_num > 0:
                    neg_scores = torch.masked_select(flat_input, neg_example.view(-1, 1).bool()).view(-1, logits_size)
                    neg_scores_idx = neg_scores[:, label_idx]
                    neg_scores_sort, _ = torch.sort(neg_scores_idx, )
                    threshold = neg_scores_sort[-keep_num + 1]
                    cond = (torch.argmax(flat_input, dim=1) == label_idx & flat_input[:, label_idx] >= threshold) | pos_example.view(
                        -1)
                    ohem_mask_idx = torch.where(cond, 1, 0)

                    flat_input_idx = flat_input[:, label_idx]
                    flat_target_idx = flat_target[:, label_idx]

                    flat_input_idx = flat_input_idx * ohem_mask_idx
                    flat_target_idx = flat_target_idx * ohem_mask_idx
                else:
                    flat_input_idx = flat_input[:, label_idx]
                    flat_target_idx = flat_target[:, label_idx]

                loss_idx = self._compute_dice_loss(flat_input_idx.view(-1, 1), flat_target_idx.view(-1, 1))
                if loss is None:
                    loss = loss_idx
                else:
                    loss += loss_idx
            return loss

        else:
            for label_idx in range(logits_size):
                flat_input_idx = flat_input[:, label_idx]
                flat_target_idx = flat_target[:, label_idx]

                loss_idx = self._compute_dice_loss(flat_input_idx.view(-1, 1), flat_target_idx.view(-1, 1))
                if loss is None:
                    loss = loss_idx
                else:
                    loss += loss_idx
            return loss

    def _binary_class(self, logits, target, mask=None):
        flat_input = logits.view(-1)
        flat_target = target.view(-1).float()
        flat_input = torch.sigmoid(flat_input) if self.with_logits else flat_input

        if mask is not None:
            mask = mask.float()
            flat_input = flat_input * mask
            flat_target = flat_target * mask
        else:
            mask = torch.ones_like(target)

        if self.ohem_ratio > 0:
            pos_example = target > 0.5
            neg_example = target <= 0.5
            mask_neg_num = mask <= 0.5

            pos_num = pos_example.sum() - (pos_example & mask_neg_num).sum()
            neg_num = neg_example.sum()
            keep_num = min(int(pos_num * self.ohem_ratio), neg_num)

            neg_scores = torch.masked_select(flat_input, neg_example.bool())
            neg_scores_sort, _ = torch.sort(neg_scores, )
            threshold = neg_scores_sort[-keep_num + 1]
            cond = (flat_input > threshold) | pos_example.view(-1)
            ohem_mask = torch.where(cond, 1, 0)
            flat_input = flat_input * ohem_mask
            flat_target = flat_target * ohem_mask

        return self._compute_dice_loss(flat_input, flat_target)

    def __str__(self):
        return f"Dice Loss smooth:{self.smooth}, ohem: {self.ohem_ratio}, alpha: {self.alpha}"

    def __repr__(self):
        return str(self)


class GlobalPointerLoss(nn.Module):
    """https://kexue.fm/archives/7359"""

    def forward(self, y_pred, y_true):
        if y_true.is_sparse:
            y_true = y_true.to_dense()
        bz, n_heads = y_pred.shape[:2]
        y_true = y_true.reshape([bz * n_heads, -1])
        y_pred = y_pred.reshape([bz * n_heads, -1])

        # multilabel_categorical_crossentory
        y_pred = (1 - 2 * y_true) * y_pred
        y_pred_neg = y_pred - y_true * 1e12
        y_pred_pos = y_pred - (1 - y_true) * 1e12
        zeros = torch.zeros_like(y_pred[..., :1])
        y_pred_neg = torch.cat([y_pred_neg, zeros], dim=-1)
        y_pred_pos = torch.cat([y_pred_pos, zeros], dim=-1)
        neg_loss = torch.logsumexp(y_pred_neg, dim=-1)
        pos_loss = torch.logsumexp(y_pred_pos, dim=-1)
        return (neg_loss + pos_loss).mean()


class RDropLoss(nn.Module):
    """https://github.com/dropreg/R-Drop"""

    def __init__(self, alpha=4):
        super().__init__()
        self.alpha = alpha
        self.ce = nn.CrossEntropyLoss(reduction='none')
        self.kld = nn.KLDivLoss(reduction='none')

    def forward(self, logits1, logits2, target):
        ce_loss = (self.ce(logits1, target) + self.ce(logits2, target)) / 2

        prob1 = F.log_softmax(logits1, dim=-1)
        prob2 = F.log_softmax(logits2, dim=-1)
        kl_loss1 = self.kld(prob1, prob2).sum(-1)
        kl_loss2 = self.kld(prob2, prob1).sum(-1)
        kl_loss = (kl_loss1 + kl_loss2) / 2
        return ce_loss + kl_loss * self.alpha


LOSSES = {
    'ce': nn.CrossEntropyLoss,
    'bce': nn.BCELoss,
    'bce-l': nn.BCEWithLogitsLoss,
    'l1': nn.L1Loss,
    'focal': FocalLoss,
    'dice': DiceLoss,
    'gp': GlobalPointerLoss,
    'r-drop': RDropLoss,
}


def get(name: str):
    assert name is not None
    loss = LOSSES.get(name)
    if loss is None:
        raise TailorsError(f"Not implemented loss: {name}")
    return loss()



# -*- coding: utf-8 -*-
import itertools
from collections import namedtuple
from dataclasses import dataclass
from typing import List, Optional, Union

import hao

Example = namedtuple('Example', ['idx', 'features', 'target'])


class Derivable:

    @classmethod
    def subclasses(cls):
        all_subclasses = []
        for subclass in cls.__subclasses__():
            all_subclasses.append(subclass)
            all_subclasses.extend(subclass.subclasses())
        return all_subclasses


class Tags:
    __slot__ = ['tag2id', 'id2tag', 'n_tags', 'supporting_tags', 'supporting_tag_ids', 'trivial_tag', 'trivial_tag_id']
    def __init__(self,
                 tags: List[str],
                 supporting_tags: Optional[List[str]] = None,
                 trivial_tag: str = 'O',
                 is_bies = False) -> None:
        supporting_tags = supporting_tags or []
        self.tag2id, self.id2tag, self.n_tags = self.build_mappings(tags, supporting_tags, trivial_tag, is_bies)
        self.supporting_tags = set(supporting_tags)
        self.supporting_tag_ids = set([self.tag2id[tag] for tag in supporting_tags])
        self.trivial_tag = trivial_tag
        self.trivial_tag_id = self.tag2id[trivial_tag]

    @staticmethod
    def build_mappings(tags: List[str],
                       supporting_tags: Optional[List[str]] = None,
                       trivial_tag: str = 'O',
                       is_bies = False):
        assert tags is not None and len(tags) > 0, 'expecting `tags` of list of str'
        if supporting_tags:
            assert all(tag not in tags for tag in supporting_tags), '`tags` should not contain any of `supporting_tags`'

        tags, supporting_tags = hao.lists.uniquify(tags), hao.lists.uniquify(supporting_tags)
        if is_bies:
            prefixes = ('B-', 'I-', 'E-', 'S-')
            tags = [f"{prefix}{tag}" for tag, prefix in itertools.product(tags, prefixes)]
        tags_all = tags + (supporting_tags or [])
        if trivial_tag not in tags_all:
            tags_all.append(trivial_tag)

        tag2id = {tag: i for i, tag in enumerate(tags_all)}
        id2tag = {i: tag for i, tag in enumerate(tags_all)}
        return tag2id, id2tag, len(tags_all)

    def get_id_by_tag(self, tag: str, default=None, ignore_supporting=False):
        if ignore_supporting and tag in self.supporting_tag:
            return self.trivial_tag_id
        return self.tag2id.get(tag, default or self.trivial_tag_id)

    def get_ids_by_tags(self, tags: List[str], default=None, ignore_supporting=False, drop_supporting=False):
        return [self.get_id_by_tag(tag, default, ignore_supporting) for tag in tags if not drop_supporting or tag not in self.supporting_tags]

    def get_tag_by_id(self, _id: int, default=None, ignore_supporting=False):
        if ignore_supporting and _id in self.supporting_tag_ids:
            return self.trivial_tag
        return self.id2tag.get(_id, default or self.trivial_tag)

    def get_tags_by_ids(self, ids: List[int], default=None, ignore_supporting=False, drop_supporting=False):
        return [self.get_tag_by_id(_id, default, ignore_supporting) for _id in ids if not drop_supporting or _id not in self.supporting_tag_ids]

    def is_supporting_tag(self, tag: str):
        return tag in self.supporting_tags

    def is_supporting_tag_id(self, _id):
        return _id in self.supporting_tag_ids

    def is_not_supporting_tag_id(self, _id):
        return not self.is_supporting_tag_id(_id)


@dataclass
class Factor:
    factor: float
    min_val: float = 1e-10
    max_val: float = 1e-1

    def apply(self, v: Union[int, float]):
        val = self.factor * v
        if self.min_val:
            val = max(self.min_val, val)
        if self.max_val:
            val = min(self.max_val, val)
        return val



# -*- coding: utf-8 -*-
from typing import Optional

import numpy as np


def divide(numerator, denominator):
    """Performs division and handles divide-by-zero. """
    mask = denominator == 0.0
    denominator = denominator.copy()
    denominator[mask] = 1  # avoid infs/nans
    result = numerator / denominator

    if not np.any(mask):
        return result

    result[mask] = 0.0
    return result


def f_score(precision, recall, beta=1.0):
    beta2 = beta ** 2
    if np.isposinf(beta):
        return recall
    else:
        denom = beta2 * precision + recall

        denom[denom == 0.] = 1  # avoid division by 0
        return (1 + beta2) * precision * recall / denom


def precision_recall_f1_support(pred_sum, tp_sum, true_sum, average: Optional[str] = None):
    weights = None
    if average == 'micro':
        tp_sum = np.array([tp_sum.sum()])
        pred_sum = np.array([pred_sum.sum()])
        true_sum = np.array([true_sum.sum()])
    elif average == 'weighted':
        weights = true_sum

    precision = divide(tp_sum, pred_sum)
    recall = divide(tp_sum, true_sum)
    f1 = f_score(precision, recall)

    if average is not None:
        precision = np.average(precision, weights=weights)
        recall = np.average(recall, weights=weights)
        f1 = np.average(f1, weights=weights)
        true_sum = sum(true_sum)
    return precision, recall, f1, true_sum



# -*- coding: utf-8 -*-
__version__ = '0.1.4'

import os
import random
from collections import UserDict
from typing import List

import hao
import torch
from decorator import decorator
from torch.nn import functional as F

LOGGER = hao.logs.get_logger(__name__)


def freeze(model):
    for param in model.parameters():
        param.requires_grad = False


def unfreeze(model):
    for param in model.parameters():
        param.requires_grad = True


def init(model):
    for layer in model.modules():
        if isinstance(layer, torch.nn.Conv2d):
            torch.nn.init.kaiming_normal_(layer.weight, mode='fan_out', nonlinearity='relu')
            if layer.bias is not None:
                torch.nn.init.constant_(layer.bias, val=0.0)

        elif isinstance(layer, torch.nn.BatchNorm2d):
            torch.nn.init.constant_(layer.weight, val=1.0)
            torch.nn.init.constant_(layer.bias, val=0.0)

        elif isinstance(layer, torch.nn.Linear):
            torch.nn.init.xavier_normal_(layer.weight)
            if layer.bias is not None:
                torch.nn.init.constant_(layer.bias, val=0.0)


def set_seed(seed, deterministic=True, benchmark=False):
    if seed is None:
        seed = random.randint(0, 10000)
        LOGGER.info(f"[seed] using random generated seed: {seed}")
    else:
        LOGGER.info(f"[seed] using random seed: {seed}")

    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    try:
        import numpy as np
        np.random.seed(seed)
    except ImportError:
        pass
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = deterministic
    torch.backends.cudnn.benchmark = benchmark


def move_to_device(data, device, non_blocking=True):
    if device is None:
        return
    if isinstance(data, (dict, UserDict)):
        return {k: move_to_device(v, device, non_blocking) for k, v in data.items()}
    elif isinstance(data, tuple):
        if hasattr(data, '_asdict') and hasattr(data, '_fields'):
            cls = type(data)
            return cls(*[move_to_device(v, device, non_blocking) for v in data])
        else:
            return tuple([move_to_device(v, device, non_blocking) for v in data])
    elif isinstance(data, list):
        return [move_to_device(v, device, non_blocking) for v in data]
    elif isinstance(data, torch.Tensor):
        return data.to(device, non_blocking=non_blocking)
    return data


@decorator
def auto_device(func, *a, **kw):
    def wrapper(self, *args, **kwargs):
        args = move_to_device(args, self.device)
        kwargs = move_to_device(kwargs, self.device)
        return func(self, *args, **kwargs)
    return wrapper(*a, **kw)


def off_tensor(data):
    if data is None:
        return None
    if isinstance(data, dict):
        return {k: off_tensor(v) for k, v in data.items()}
    elif isinstance(data, tuple):
        return tuple([off_tensor(v) for v in data])
    elif isinstance(data, list):
        return [off_tensor(v) for v in data]
    elif isinstance(data, torch.Tensor):
        return data.detach().cpu() if data.is_sparse else data.tolist()
    return data


def append_batch(items, batch):
    batch = off_tensor(batch)
    if isinstance(batch, tuple):
        hao.lists.add_tuple_to_list(items, batch)
    elif isinstance(batch, torch.Tensor):
        items.append(batch)
    else:
        items.extend(batch)


def multi_hot(items: List[int], n_classes: int, dtype=torch.float32):
    data = torch.LongTensor(items)
    onehot = F.one_hot(data, num_classes=n_classes)
    return onehot.sum(dim=0).to(dtype)



# -*- coding: utf-8 -*-
import functools
from typing import List, Optional

import torch
from sklearn import metrics

precision_score = functools.partial(metrics.precision_score, average='micro', zero_division=0)
recall_score = functools.partial(metrics.recall_score, average='micro', zero_division=0)
f1_score = functools.partial(metrics.f1_score, average='micro', zero_division=0)
precision_recall_fscore_support = functools.partial(metrics.precision_recall_fscore_support, average='micro', zero_division=0)
classification_report = functools.partial(metrics.classification_report, digits=4, zero_division=0)


def calculate(target, preds, labels: Optional[List[str]] = None):
    if isinstance(target, torch.Tensor):
        targets = target.tolist()
    precision, recall, f1, _ = precision_recall_fscore_support(targets, preds, labels=labels)
    report = classification_report(targets, preds, labels=labels)
    return {'precision': precision, 'recall': recall, 'f1': f1, 'report': report}



# -*- coding: utf-8 -*-
import abc
import tempfile

import hao
from transformers import (BertModel, BertTokenizerFast, ElectraModel,
                          ElectraTokenizerFast, PreTrainedModel,
                          PreTrainedTokenizerFast, XLNetModel,
                          XLNetTokenizerFast)

from tailors import freeze
from tailors.domains import Derivable


class PLM(abc.ABC, Derivable):
    def __init__(self, model_conf) -> None:
        super().__init__(model_conf)
        self.model_conf = model_conf
        self.tokenizer = self.build_tokenizer()
        self.plm = self.build_plm()

    @property
    def config(self):
        return self.plm.config

    def get_additional_tokens_for_tokenizer(self):
        return []

    @abc.abstractmethod
    def build_tokenizer(self) -> PreTrainedTokenizerFast:
        raise NotImplementedError()

    @abc.abstractmethod
    def build_plm(self) -> PreTrainedModel:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_embedding_size(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_save_checkpoint(self, checkpoint) -> None:
        raise NotImplementedError()

    @staticmethod
    def get_pretrained_model_path(model: str):
        # hao.oss.init(f"pretrained.{model}", config='tailors.yml')
        key = f"pretrained.{model}.local"
        return hao.config.get_path(key) or hao.config.get_path(key, config='tailors.yml')


class Bert(PLM):

    def build_tokenizer(self):
        if hasattr(self.model_conf, 'vocab'):
            with tempfile.NamedTemporaryFile() as f:
                f.write(self.model_conf.vocab.encode())
                tokenizer = BertTokenizerFast(vocab_file=f.name)
                if hasattr(self.model_conf, 'additional_special_tokens'):
                    tokenizer.add_special_tokens({'additional_special_tokens': self.model_conf.additional_special_tokens})
                return tokenizer
        else:
            pretrained_model_path = self.get_pretrained_model_path('bert')
            tokenizer = BertTokenizerFast.from_pretrained(pretrained_model_path)
            additional_special_tokens = self.get_additional_tokens_for_tokenizer()
            if additional_special_tokens is not None and len(additional_special_tokens) > 0:
                tokenizer.add_special_tokens({'additional_special_tokens': additional_special_tokens})
            return tokenizer

    def build_plm(self):
        if hasattr(self.model_conf, 'bert_config'):
            bert = BertModel(self.model_conf.bert_config)
        else:
            pretrained_model_path = self.get_pretrained_model_path('bert')
            bert = BertModel.from_pretrained(pretrained_model_path)
        if self.model_conf.freeze_embedding:
            freeze(bert)
        return bert

    def get_embedding_size(self):
        return self.plm.config.hidden_size

    def on_save_checkpoint(self) -> None:
        if not hasattr(self.model_conf, 'vocab'):
            pretrained_model_path = self.get_pretrained_model_path('bert')
            path_vocab_file = hao.paths.get_path(pretrained_model_path, 'vocab.txt')
            self.model_conf.vocab = open(path_vocab_file).read()
        if len(self.tokenizer.additional_special_tokens) > 0:
            self.model_conf.additional_special_tokens = self.tokenizer.additional_special_tokens
        self.model_conf.bert_config = self.plm.config


class XLNet(PLM):

    def build_tokenizer(self):
        if hasattr(self.model_conf, 'vocab'):
            with tempfile.NamedTemporaryFile() as f:
                f.write(self.model_conf.vocab)
                tokenizer = XLNetTokenizerFast(vocab_file=f.name)
                if hasattr(self.model_conf, 'additional_special_tokens'):
                    tokenizer.add_special_tokens({'additional_special_tokens': self.model_conf.additional_special_tokens})
                return tokenizer
        else:
            pretrained_model_path = self.get_pretrained_model_path('xlnet')
            tokenizer = XLNetTokenizerFast.from_pretrained(pretrained_model_path)
            additional_special_tokens = self.get_additional_tokens_for_tokenizer()
            if additional_special_tokens is not None and len(additional_special_tokens) > 0:
                tokenizer.add_special_tokens({'additional_special_tokens': additional_special_tokens})
            return tokenizer

    def build_plm(self):
        if hasattr(self.model_conf, 'xlnet_config'):
            xlnet = XLNetModel(self.model_conf.xlnet_config)
        else:
            pretrained_model_path = self.get_pretrained_model_path('xlnet')
            xlnet = XLNetModel.from_pretrained(pretrained_model_path)
        if self.model_conf.freeze_embedding:
            freeze(xlnet)
        return xlnet

    def get_embedding_size(self):
        return self.plm.config.d_model

    def on_save_checkpoint(self) -> None:
        if not hasattr(self.model_conf, 'vocab'):
            pretrained_model_path = self.get_pretrained_model_path('xlnet')
            path_vocab_file = hao.paths.get_path(pretrained_model_path, 'spiece.model')
            self.model_conf.vocab = open(path_vocab_file, 'rb').read()
        if len(self.tokenizer.additional_special_tokens) > 0:
            self.model_conf.additional_special_tokens = self.tokenizer.additional_special_tokens
        self.model_conf.xlnet_config = self.plm.config


class Electra(PLM):

    def build_tokenizer(self):
        if hasattr(self.model_conf, 'vocab'):
            with tempfile.NamedTemporaryFile() as f:
                f.write(self.model_conf.vocab.encode())
                tokenizer = ElectraTokenizerFast(vocab_file=f.name)
                if hasattr(self.model_conf, 'additional_special_tokens'):
                    tokenizer.add_special_tokens({'additional_special_tokens': self.model_conf.additional_special_tokens})
                return tokenizer
        else:
            pretrained_model_path = self.get_pretrained_model_path('electra')
            tokenizer = ElectraTokenizerFast.from_pretrained(pretrained_model_path)
            additional_special_tokens = self.get_additional_tokens_for_tokenizer()
            if additional_special_tokens is not None and len(additional_special_tokens) > 0:
                tokenizer.add_special_tokens({'additional_special_tokens': additional_special_tokens})
            return tokenizer

    def build_plm(self):
        if hasattr(self.model_conf, 'electra_config'):
            electra = ElectraModel(self.model_conf.electra_config)
        else:
            pretrained_model_path = self.get_pretrained_model_path('electra')
            electra = ElectraModel.from_pretrained(pretrained_model_path)
        if self.model_conf.freeze_embedding:
            freeze(electra)
        return electra

    def get_embedding_size(self):
        return self.plm.config.hidden_size

    def on_save_checkpoint(self) -> None:
        if not hasattr(self.model_conf, 'vocab'):
            pretrained_model_path = self.get_pretrained_model_path('electra')
            path_vocab_file = hao.paths.get_path(pretrained_model_path, 'vocab.txt')
            self.model_conf.vocab = open(path_vocab_file).read()
        if len(self.tokenizer.additional_special_tokens) > 0:
            self.model_conf.additional_special_tokens = self.tokenizer.additional_special_tokens
        self.model_conf.electra_config = self.plm.config


def list_plms():
    return PLM.subclasses()
